//------------------------------------------------------------------------
//
// OpenDCC - S88-N - Keyboard and RC5 Receiver
//
// Copyright (c) 2007 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      s88_key.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2007-11-27 V0.1 start
//            2008-01-10 V0.2 keyboard tested and okay (both modes)
//
//------------------------------------------------------------------------
//
// purpose:   keyboard routines for directly connected keys
//
// content:   Here we scan the inputs and do the debounce.
//
//            Every Input has a holdtime - this will prolong
//            each stroke to a given time.
//
//            Currently we support two types of keyboards
//              1: standard direct keys to ground.
 //             2: Pollin Dual Scan (Samwell with 12 Keys) 
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "main.h"                // timerval, current_message
//------------------------------------------------------------------------------

#define KEY_SCAN_TICK    10000L    // scan keyboard every 10ms
#define KEY_DEBOUNCE     (20000L / KEY_SCAN_TICK)        // 20ms Entprellen
#define HOLDTIME         (1000000L / KEY_SCAN_TICK)      // Hold 1s

#define nop() __asm volatile ("nop")

//==============================================================================
//
#define NUM_OF_KEYS      16    // 12

struct
  {
    enum                  
      {                           
        S_UNPRESSED,
        S_PRESS_DEBOUNCE,
        S_PRESSED,
        S_RELEASE_DEBOUNCE
      } state; 
    unsigned char holdtime;   // time to keep the pressed state (given in ticks (20ms)
    unsigned char debounce;   // debounce in both directions
  } key[NUM_OF_KEYS];

//
//=========================================================================
// KEYBOARD TYPE 1
//-------------------------------------------------------------------------
// This is for standard keys connected to ground
//
// Connect the Attiny2313 as follows:
//
//
//                       --------U--------
//       (reset)     NC  | 1 PA2   VCC 20|-
//   +--|>|-- 2k -- LED  | 2 PD0   PB7 19|- S88_CLK
//            S88_RESET  | 3 PD1   PB6 18|- S88_DO
//       (xtal) JUMPER2  | 4 PA1   PB5 17|- S88_DI
//       (xtal) JUMPER1  | 5 PA0   PB4 16|- S88_LOAD
//         RC5_receiver  | 6 PD2   PB3 15|- KEY 8
//                 KEY1  | 7 PD3   PB2 14|- KEY 7
//                 KEY2  | 8 PD4   PB1 13|- KEY 6
//                 KEY3  | 9 PD5   PB0 12|- KEY 5
//                  GND  |10 GND   PD6 11|- KEY 4
//                       -----------------
//
// cable wiring:
//              --------
//         GND  |1    2|  Key 1
//         GND  |3    4|  Key 2
//         GND  |5    6|  Key 3
//         GND  |7    8|  Key 4
//         GND  |9   10|  Key 5
//         GND  |11  12|  Key 6
//         GND  |13  14|  Key 7
//         GND  |15  16|  Key 8
//              --------
//
//----------------------------------------------------------------------------


unsigned char rearrange_std[8] = 
  {
    //76543210          // keys in Stream
    //||||----
    //                  // index=scanned bit (starting from LSB)
    //                  // content: target position
    0b10000000,         // 
    0b01000000,         // 
    0b00100000,         // 
    0b00010000,         // 
    0b00001000,         //
    0b00000100,         //
    0b00000010,         //
    0b00000001,         // 
  };

// --- the real hardware interface - get 8 bit of data.
unsigned char read_keys_standard(void)
  {
    unsigned char read_i, result, temp;

    read_i = PINB & 0b00001111;
    read_i = read_i << 4;
    temp = PIND & 0b01111000;
    temp = temp >> 3;
    read_i = read_i | temp;
    read_i = ~read_i;            // key strokes are active low -> we deliver active '1'

    // flip vector - s88 is reading bit 0 from the msb

    result = 0;
    for (temp = 0; temp<(sizeof(rearrange_std)/sizeof(rearrange_std[0])); temp++)
      {
        if (read_i & (1<<temp))
          result |= rearrange_std[temp];
      }
    return(result);    
  }


//=========================================================================
// KEYBOARD TYPE 2
//-------------------------------------------------------------------------
//
// Connect the Attiny2313 as follows:
// (tp__ = pin at pollin keyboard SAMWELL 702772-8)
// for SAMWELL keyboard: D2, D3 scan output
//
//                             --------U--------
//             (reset)     NC  | 1 PA2   VCC 20|-
//        + --|>|-- 2k -- LED  | 2 PD0   PB7 19|- S88_CLK
//                  S88_RESET  | 3 PD1   PB6 18|- S88_DO
//             (xtal) JUMPER2  | 4 PA1   PB5 17|- S88_DI
//             (xtal) JUMPER1  | 5 PA0   PB4 16|- S88_LOAD
//  (RC5_receiver) COL1 (tp01) | 6 PD2   PB3 15|- (tp12) ROW7 (KEY8)
//  (KEY1)         COL2 (tp02) | 7 PD3   PB2 14|- (tp11) ROW6 (KEY7)
//  (KEY2)         ROW1 (tp06) | 8 PD4   PB1 13|- (tp10) ROW5 (KEY6)
//  (KEY3)         ROW2 (tp07) | 9 PD5   PB0 12|- (tp09) ROW4 (KEY5)
//                        GND  |10 GND   PD6 11|- (tp08) ROW3 (KEY4)
//                             -----------------
// COL1 and COL2 are fed low during scan.
//
// cable wiring:          Samwell:
//              --------
//         GND  |1    2|  Pin 2
//         GND  |3    4|  Pin 6
//         GND  |5    6|  Pin 7
//         GND  |7    8|  Pin 8
//         GND  |9   10|  Pin 9
//         GND  |11  12|  Pin 10
//         GND  |13  14|  Pin 11
//         GND  |15  16|  Pin 12
//              --------
// extra wire: RC5 Pad <-> Pin 1
//--------------------------------------------------------------------------

unsigned int rearrange_2col[16] = 
  {
    //1234567890v^....          // keys in S88-Stream
    //                          // index=scanned bit (starting from LSB)
    //                          // content: target position of this keystroke in S88
    //||||----||||----          //  
    0b0000010000000000,         // Col 1, PB0 -> 6
    0b0000000100000000,         // Col 1, PB1 -> 8
    0b0000000000010000,         // Col 1, PB2 -> C^
    0b0000000000100000,         // Col 1, PB3 -> Cv
    0b0000000001000000,         // Col 1, PD4 -> 0
    0b0100000000000000,         // Col 1, PD5 -> 2
    0b0001000000000000,         // Col 1, PD6 -> 4
    0b0000000000000000,         // Col 1, xxx -> empty    
    //||||----||||----          //  
    0b0000001000000000,         // Col 2, PB0 -> 7
    0b0000000010000000,         // Col 2, PB1 -> 9
    0b0000000000000000,         // Col 2, PB2 -> empty
    0b0000000000000000,         // Col 2, PB3 -> empty
    0b1000000000000000,         // Col 2, PD4 -> 1
    0b0010000000000000,         // Col 2, PD5 -> 3
    0b0000100000000000,         // Col 2, PD6 -> 5
    0b0000000000000000,         // Col 2, xxx -> empty  
  };

// --- the real hardware interface - get 12 bit of data.
unsigned int read_keys_2col_pollin(void)
  {
    unsigned int read_i;
    unsigned int result;
    unsigned char temp1, temp2;

    // scan COL 2 
    DDRD |= (1<<PD3);     // output
    PORTD &= ~(1<<PD3);   // Pull this line low
    nop();nop();nop();
    temp1 = PINB & 0b00001111;
    temp2 = PIND & 0b01110000;
    temp2 = temp2 | temp1;
    read_i = temp2 * 256;
    PORTD |= (1<<PD3);
    DDRD &= ~(1<<PD3);     // input
    
      
    // scan COL 1
    DDRD |= (1<<PD2);     // output
    PORTD &= ~(1<<PD2);   // Pull this line low
    nop();nop();nop();
    temp1 = PINB & 0b00001111;
    temp2 = PIND & 0b01110000;
    temp2 = temp2 | temp1;
    read_i = read_i | temp2;
    PORTD |= (1<<PD2);
    DDRD &= ~(1<<PD2);     // input
    
    read_i = ~read_i;     // key strokes are active low -> we deliver active '1'
    
    // rearrange bits
    result = 0;
    for (temp1 = 0; temp1<(sizeof(rearrange_2col)/sizeof(rearrange_2col[0])); temp1++)
      {
        if (read_i & (1<<temp1))
          result |= rearrange_2col[temp1];
      }
    return(result);      
  }

//========================================================================================
//
//  KEYBOARD - Handler
//
//========================================================================================

unsigned char keyboard_type;

void init_keyboard(void)
  {
    unsigned char i;

    if (JUMPER1_FITTED) keyboard_type = 2;          // then pollin 
    else keyboard_type = 1;                         // else std

    for (i=0; i<NUM_OF_KEYS; i++)
      {
        key[i].state = S_UNPRESSED;
        key[i].holdtime = 0;
      }
  }


void update_keys(void)
  {
    #if (NUM_OF_KEYS <= 8)
      unsigned char mask;
      unsigned char pattern;
    #elif (NUM_OF_KEYS <= 16)
      unsigned int mask;
      unsigned int pattern;
    #else
      #warning NUM_OF_KEYS too large
    #endif
    
    unsigned char i;

    switch (keyboard_type)
      {
        default:
        case 0: pattern = 0;
                break;
        case 1: pattern = read_keys_standard();             // fetch actual state
                break;
        case 2: pattern = read_keys_2col_pollin();          // fetch actual state
                break;
      }

    current_message = 0;                // auto reset
    mask = 1;
    for (i=0; i<NUM_OF_KEYS; i++)
      {
        switch(key[i].state)
          {
            case S_UNPRESSED:
                if ((pattern & mask))
                  {
                    // got pressed
                    key[i].debounce = KEY_DEBOUNCE;
                    key[i].state = S_PRESS_DEBOUNCE;
                  }
                break;
            case S_PRESS_DEBOUNCE:
                if ((pattern & mask))
                  {
                    // still pressed
                    key[i].debounce--;
                    if (key[i].debounce == 0) key[i].state = S_PRESSED;
                  }
                else
                  {
                    key[i].state = S_UNPRESSED;
                  }
                break;
            case S_PRESSED:
                if ((pattern & mask))
                  {
                    // still pressed
                    key[i].holdtime = HOLDTIME;
                  }
                else
                  {
                    // got released
                    key[i].debounce = KEY_DEBOUNCE;
                    key[i].state = S_RELEASE_DEBOUNCE;
                  }
                break;
            case S_RELEASE_DEBOUNCE:
                if ((pattern & mask))
                  {
                    // still pressed - go back 
                    key[i].state = S_PRESSED;
                  }
                else
                  {
                    
                    key[i].debounce--;
                    if (key[i].debounce == 0) key[i].state = S_UNPRESSED;
                  }
                break;
          }

        if (key[i].holdtime > 0)                // whoever has a holdtime > 0: is active
          {
            key[i].holdtime--;
            current_message |= mask;
          }
        mask = mask << 1;
      }
  }


